@@ -156,6 +156,37 @@ class MonitoringConfig:
         return self.is_enabled(environment)
 
 
+@dataclass
+class MigrationConfig:
+    """Migration and schema management configuration.
+
+    Controls schema cache behavior, migration operations, and failure handling.
+    See ADR-001 for comprehensive architecture details.
+    """
+
+    # Migration behavior
+    auto_migrate: bool = True
+    migration_enabled: bool = True
+    existing_schema_mode: bool = False
+
+    # Schema cache settings (ADR-001)
+    schema_cache_enabled: bool = True
+    schema_cache_ttl: Optional[int] = None  # None = no expiration, or seconds
+    schema_cache_max_size: int = 10000
+    schema_cache_validation: bool = False  # Enable schema checksum validation
+
+    # Failure handling (exponential backoff)
+    migration_retry_enabled: bool = True
+    migration_max_failures: int = 3
+    migration_failure_backoff: int = 60  # seconds
+
+    # Locking
+    migration_lock_timeout: int = 30  # seconds
+
+    # Observability
+    migration_metrics_enabled: bool = True
+
+
 @dataclass
 class SecurityConfig:
     """Security configuration with enterprise defaults"""
@@ -260,13 +291,51 @@ class DataFlowConfig:
                 sec_config_kwargs["audit_enabled"] = kwargs["audit_logging"]
             self.security = SecurityConfig(**sec_config_kwargs)
 
+        # Handle migration configuration (ADR-001)
+        if "migration" in kwargs and isinstance(kwargs["migration"], MigrationConfig):
+            self.migration = kwargs["migration"]
+        else:
+            migration_config_kwargs = {}
+            # Backward compatibility with existing parameters
+            if "auto_migrate" in kwargs:
+                migration_config_kwargs["auto_migrate"] = kwargs["auto_migrate"]
+            if "existing_schema_mode" in kwargs:
+                migration_config_kwargs["existing_schema_mode"] = kwargs[
+                    "existing_schema_mode"
+                ]
+            # Schema cache parameters (ADR-001)
+            if "schema_cache_enabled" in kwargs:
+                migration_config_kwargs["schema_cache_enabled"] = kwargs[
+                    "schema_cache_enabled"
+                ]
+            if "schema_cache_ttl" in kwargs:
+                migration_config_kwargs["schema_cache_ttl"] = kwargs["schema_cache_ttl"]
+            if "schema_cache_max_size" in kwargs:
+                migration_config_kwargs["schema_cache_max_size"] = kwargs[
+                    "schema_cache_max_size"
+                ]
+            if "schema_cache_validation" in kwargs:
+                migration_config_kwargs["schema_cache_validation"] = kwargs[
+                    "schema_cache_validation"
+                ]
+            # Failure handling parameters
+            if "migration_max_failures" in kwargs:
+                migration_config_kwargs["migration_max_failures"] = kwargs[
+                    "migration_max_failures"
+                ]
+            if "migration_failure_backoff" in kwargs:
+                migration_config_kwargs["migration_failure_backoff"] = kwargs[
+                    "migration_failure_backoff"
+                ]
+            self.migration = MigrationConfig(**migration_config_kwargs)
+
         # Node generation settings
         self.auto_generate_nodes = kwargs.get("auto_generate_nodes", True)
         self.node_prefix = kwargs.get("node_prefix", "")
         self.node_suffix = kwargs.get("node_suffix", "Node")
 
-        # Migration settings
-        self.auto_migrate = kwargs.get("auto_migrate", True)
+        # Backward compatibility: Keep auto_migrate as direct property
+        self.auto_migrate = self.migration.auto_migrate
         self.migration_directory = Path(kwargs.get("migration_directory", "migrations"))
 
         # Cache settings
