@@ -263,9 +263,7 @@ class BulkOperations:
 
         # Perform actual database insertion
         try:
-            connection_string = self.dataflow.config.database.get_connection_url(
-                self.dataflow.config.environment
-            )
+            connection_string = self.dataflow.config.database.url or ":memory:"
             database_type = self.dataflow._detect_database_type()
             table_name = self.dataflow._class_name_to_table_name(model_name)
 
@@ -326,34 +324,48 @@ class BulkOperations:
                     f"BULK_CREATE: Executing batch {batches_processed + 1}, query='{query[:100]}...', param_count={len(params)}"
                 )
 
-                # Execute using AsyncSQLDatabaseNode
-                from kailash.nodes.data.async_sql import AsyncSQLDatabaseNode
+                # Execute using cached AsyncSQLDatabaseNode
+                # FIX: Use cached node instead of creating fresh instance
+                # This ensures connection pooling and data visibility across operations
+                sql_node = self.dataflow._get_or_create_async_sql_node(database_type)
 
-                sql_node = AsyncSQLDatabaseNode(
-                    connection_string=connection_string,
-                    database_type=database_type,
+                result = await sql_node.async_run(
                     query=query,
                     params=params,
                     fetch_mode="all",
                     validate_queries=False,
                     transaction_mode="auto",
                 )
-                result = await sql_node.async_run()
 
                 logger.warning(f"BULK_CREATE: SQL result={result}")
 
                 # Extract rows_affected from result
-                # For INSERT: AsyncSQLDatabaseNode returns row_count at result level
-                # For DELETE: returns rows_affected in data[0]
+                # Handle multiple formats (check data field FIRST for accurate rowcount):
+                # - New: data is dict with rows_affected: {"rows_affected": N}
+                # - New: data is list with rows_affected: [{"rows_affected": N}]
+                # - Old: row_count at result level (fallback)
                 rows_affected = 0
                 if result and "result" in result:
                     result_data = result["result"]
-                    # Try row_count first (INSERT operations)
-                    if "row_count" in result_data:
+                    # Try data field first (new DML format - most accurate)
+                    if "data" in result_data:
+                        data = result_data["data"]
+                        # Handle dict format: {"rows_affected": N}
+                        if isinstance(data, dict) and "rows_affected" in data:
+                            rows_affected = data["rows_affected"]
+                        # Handle list format: [{"rows_affected": N}]
+                        elif (
+                            isinstance(data, list)
+                            and len(data) > 0
+                            and "rows_affected" in data[0]
+                        ):
+                            rows_affected = data[0].get("rows_affected", 0)
+                        # Fall back to row_count if data doesn't have rows_affected
+                        elif "row_count" in result_data:
+                            rows_affected = result_data.get("row_count", 0)
+                    # Fall back to row_count (old INSERT format)
+                    elif "row_count" in result_data:
                         rows_affected = result_data.get("row_count", 0)
-                    # Fall back to rows_affected in data (DELETE operations)
-                    elif "data" in result_data and len(result_data["data"]) > 0:
-                        rows_affected = result_data["data"][0].get("rows_affected", 0)
 
                 total_inserted += rows_affected
                 batches_processed += 1
@@ -438,9 +450,7 @@ class BulkOperations:
                 )
 
                 # Get database connection and execute UPDATE
-                connection_string = self.dataflow.config.database.get_connection_url(
-                    self.dataflow.config.environment
-                )
+                connection_string = self.dataflow.config.database.url or ":memory:"
                 database_type = self.dataflow._detect_database_type()
                 table_name = self.dataflow._class_name_to_table_name(model_name)
 
@@ -477,29 +487,35 @@ class BulkOperations:
                     f"BULK_UPDATE: Executing query='{query}' with params={params}"
                 )
 
-                # Execute using AsyncSQLDatabaseNode
-                from kailash.nodes.data.async_sql import AsyncSQLDatabaseNode
+                # Execute using cached AsyncSQLDatabaseNode
+                # FIX: Use cached node instead of creating fresh instance
+                # This ensures connection pooling and data visibility across operations
+                sql_node = self.dataflow._get_or_create_async_sql_node(database_type)
 
-                sql_node = AsyncSQLDatabaseNode(
-                    connection_string=connection_string,
-                    database_type=database_type,
+                result = await sql_node.async_run(
                     query=query,
                     params=params,
                     fetch_mode="all",
                     validate_queries=False,
                     transaction_mode="auto",
                 )
-                result = await sql_node.async_run()
 
                 logger.warning(f"BULK_UPDATE: SQL result={result}")
 
                 # Extract rows_affected from result
-                # AsyncSQLDatabaseNode returns: {'result': {'data': [{'rows_affected': N}], ...}}
+                # NEW format: {'result': {'data': {'rows_affected': N}, ...}}
+                # OLD format: {'result': {'data': [{'rows_affected': N}], ...}}
                 rows_affected = 0
                 if result and "result" in result:
                     result_data = result["result"]
-                    if "data" in result_data and len(result_data["data"]) > 0:
-                        rows_affected = result_data["data"][0].get("rows_affected", 0)
+                    if "data" in result_data:
+                        data = result_data["data"]
+                        # Handle new format (dict with rows_affected)
+                        if isinstance(data, dict) and "rows_affected" in data:
+                            rows_affected = data["rows_affected"]
+                        # Handle old format (list with rows_affected in first item)
+                        elif isinstance(data, list) and len(data) > 0:
+                            rows_affected = data[0].get("rows_affected", 0)
 
                 success_result = {
                     "filter": filter_criteria,
@@ -605,28 +621,35 @@ class BulkOperations:
 
                         query = f"UPDATE {table_name} {set_clause} {where_clause}"
 
-                        # Execute using AsyncSQLDatabaseNode
-                        from kailash.nodes.data.async_sql import AsyncSQLDatabaseNode
+                        # Execute using cached AsyncSQLDatabaseNode
+                        # FIX: Use cached node instead of creating fresh instance
+                        # This ensures connection pooling and data visibility across operations
+                        sql_node = self.dataflow._get_or_create_async_sql_node(
+                            database_type
+                        )
 
-                        sql_node = AsyncSQLDatabaseNode(
-                            connection_string=connection_string,
-                            database_type=database_type,
+                        result = await sql_node.async_run(
                             query=query,
                             params=params,
                             fetch_mode="all",
                             validate_queries=False,
                             transaction_mode="auto",
                         )
-                        result = await sql_node.async_run()
 
                         # Extract rows_affected
+                        # NEW format: {'result': {'data': {'rows_affected': N}, ...}}
+                        # OLD format: {'result': {'data': [{'rows_affected': N}], ...}}
                         rows_affected = 0
                         if result and "result" in result:
                             result_data = result["result"]
-                            if "data" in result_data and len(result_data["data"]) > 0:
-                                rows_affected = result_data["data"][0].get(
-                                    "rows_affected", 0
-                                )
+                            if "data" in result_data:
+                                data = result_data["data"]
+                                # Handle new format (dict with rows_affected)
+                                if isinstance(data, dict) and "rows_affected" in data:
+                                    rows_affected = data["rows_affected"]
+                                # Handle old format (list with rows_affected in first item)
+                                elif isinstance(data, list) and len(data) > 0:
+                                    rows_affected = data[0].get("rows_affected", 0)
 
                         total_updated += rows_affected
 
@@ -699,9 +722,7 @@ class BulkOperations:
             logger.warning("BULK_DELETE: Processing filter-based delete")
             try:
                 # Get database connection and execute DELETE
-                connection_string = self.dataflow.config.database.get_connection_url(
-                    self.dataflow.config.environment
-                )
+                connection_string = self.dataflow.config.database.url or ":memory:"
                 database_type = self.dataflow._detect_database_type()
                 table_name = self.dataflow._class_name_to_table_name(model_name)
 
@@ -714,34 +735,58 @@ class BulkOperations:
                     filter_criteria, database_type, params_offset=0
                 )
 
+                # Execute using cached AsyncSQLDatabaseNode
+                # FIX: Use cached node instead of creating fresh instance
+                # This ensures connection pooling and data visibility across operations
+                sql_node = self.dataflow._get_or_create_async_sql_node(database_type)
+
+                logger.warning(
+                    f"BULK_DELETE: Using cached sql_node={id(sql_node)}, "
+                    f"cache_size={len(self.dataflow._async_sql_node_cache)}"
+                )
+
+                # DEBUG: First, check if records exist
+                check_query = (
+                    f"SELECT COUNT(*) as count FROM {table_name} {where_clause}"
+                )
+                check_result = await sql_node.async_run(
+                    query=check_query,
+                    params=params,
+                    fetch_mode="all",
+                    validate_queries=False,
+                    transaction_mode="auto",
+                )
+                logger.warning(f"BULK_DELETE: Pre-delete count check: {check_result}")
+
                 query = f"DELETE FROM {table_name} {where_clause}"
                 logger.warning(
                     f"BULK_DELETE: Executing query='{query}' with params={params}"
                 )
 
-                # Execute using AsyncSQLDatabaseNode
-                from kailash.nodes.data.async_sql import AsyncSQLDatabaseNode
-
-                sql_node = AsyncSQLDatabaseNode(
-                    connection_string=connection_string,
-                    database_type=database_type,
+                result = await sql_node.async_run(
                     query=query,
                     params=params,
                     fetch_mode="all",
                     validate_queries=False,
                     transaction_mode="auto",
                 )
-                result = await sql_node.async_run()
 
                 logger.warning(f"BULK_DELETE: SQL result={result}")
 
                 # Extract rows_affected from result
-                # AsyncSQLDatabaseNode returns: {'result': {'data': [{'rows_affected': N}], ...}}
+                # NEW format: {'result': {'data': {'rows_affected': N}, ...}}
+                # OLD format: {'result': {'data': [{'rows_affected': N}], ...}}
                 rows_affected = 0
                 if result and "result" in result:
                     result_data = result["result"]
-                    if "data" in result_data and len(result_data["data"]) > 0:
-                        rows_affected = result_data["data"][0].get("rows_affected", 0)
+                    if "data" in result_data:
+                        data = result_data["data"]
+                        # Handle new format (dict with rows_affected)
+                        if isinstance(data, dict) and "rows_affected" in data:
+                            rows_affected = data["rows_affected"]
+                        # Handle old format (list with rows_affected in first item)
+                        elif isinstance(data, list) and len(data) > 0:
+                            rows_affected = data[0].get("rows_affected", 0)
 
                 success_result = {
                     "filter": filter_criteria,
@@ -850,9 +895,7 @@ class BulkOperations:
 
         # Perform actual database upsert
         try:
-            connection_string = self.dataflow.config.database.get_connection_url(
-                self.dataflow.config.environment
-            )
+            connection_string = self.dataflow.config.database.url or ":memory:"
             database_type = self.dataflow._detect_database_type()
             table_name = self.dataflow._class_name_to_table_name(model_name)
 
@@ -897,19 +940,18 @@ class BulkOperations:
                     f"query='{query[:200]}...', param_count={len(params)}"
                 )
 
-                # Execute using AsyncSQLDatabaseNode
-                from kailash.nodes.data.async_sql import AsyncSQLDatabaseNode
+                # Execute using cached AsyncSQLDatabaseNode
+                # FIX: Use cached node instead of creating fresh instance
+                # This ensures connection pooling and data visibility across operations
+                sql_node = self.dataflow._get_or_create_async_sql_node(database_type)
 
-                sql_node = AsyncSQLDatabaseNode(
-                    connection_string=connection_string,
-                    database_type=database_type,
+                result = await sql_node.async_run(
                     query=query,
                     params=params,
                     fetch_mode="all",
                     validate_queries=False,
                     transaction_mode="auto",
                 )
-                result = await sql_node.async_run()
 
                 logger.warning(f"BULK_UPSERT: SQL result={result}")
 
